from .aliases import Alias, Aliases
from .basemodel import BaseModel
from .claims import Claim, Claims
from .descriptions import Descriptions
from .forms import Form, Forms
from .labels import Labels
from .language_values import LanguageValue, LanguageValues
from .lemmas import Lemmas
from .qualifiers import Qualifiers
from .references import Reference, References
from .senses import Glosses, Sense, Senses
from .sitelinks import Sitelink, Sitelinks
from .snaks import Snak, Snaks
